# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2022/2/24 
# @Function: 非极大值抑制
import numpy as np


def multiclass_nms(boxes, num_classes, match_threshold=0.6, match_metric='iou'):
    """
    :param boxes: [[cls, scores, x_min, y_min, x_max, y_max]....]
    :param num_classes:
    :param match_threshold: overlap thresh for match metric.
    :param match_metric: 'iou' or 'ios'
    """
    final_boxes = []
    for c in range(num_classes):
        id_xs = boxes[:, 0] == c
        if np.count_nonzero(id_xs) == 0:
            continue
        r = nms_new(boxes[id_xs, 1:], match_threshold, match_metric)
        final_boxes.extend(np.concatenate([np.full((r.shape[0], 1), c), r], 1))
    return final_boxes


def nms_new(det_boxes, match_threshold=0.6, match_metric='iou'):
    """
    Apply NMS to avoid detecting too many overlapping bounding boxes.
    :param det_boxes: [[scores, x_min, y_min, x_max, y_max]....]
    :param match_threshold: overlap thresh for match metric.
    :param match_metric: 'iou' or 'ios'
    """
    if det_boxes.shape[0] == 0:
        return det_boxes[[], :]
    scores = det_boxes[:, 0]
    x1 = det_boxes[:, 1]
    y1 = det_boxes[:, 2]
    x2 = det_boxes[:, 3]
    y2 = det_boxes[:, 4]
    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    n_det_boxes = det_boxes.shape[0]
    suppressed = np.zeros(n_det_boxes, dtype=np.int32)

    for _i in range(n_det_boxes):
        i = order[_i]
        if suppressed[i] == 1:
            continue
        ix1 = x1[i]
        iy1 = y1[i]
        ix2 = x2[i]
        iy2 = y2[i]
        i_area = areas[i]
        for _j in range(_i + 1, n_det_boxes):
            j = order[_j]
            if suppressed[j] == 1:
                continue
            xx1 = max(ix1, x1[j])
            yy1 = max(iy1, y1[j])
            xx2 = min(ix2, x2[j])
            yy2 = min(iy2, y2[j])
            w = max(0.0, xx2 - xx1 + 1)
            h = max(0.0, yy2 - yy1 + 1)
            inter = w * h
            if match_metric == 'iou':
                union = i_area + areas[j] - inter
                match_value = inter / union
            elif match_metric == 'ios':
                smaller = min(i_area, areas[j])
                match_value = inter / smaller
            else:
                raise ValueError()
            if match_value >= match_threshold:
                suppressed[j] = 1
    keep = np.where(suppressed == 0)[0]
    det_boxes = det_boxes[keep, :]
    return det_boxes


if __name__ == "__main__":
    boxes_ = np.array([[1, 0.72, 100, 100, 210, 210],
                      [1, 0.80, 250, 250, 420, 420],
                      [1, 0.92, 220, 220, 320, 330],
                      [2, 0.72, 100, 100, 210, 210],
                      [2, 0.81, 230, 240, 325, 330],
                      [2, 0.90, 220, 230, 315, 340]], dtype=np.float16)
    print(boxes_.shape)
    res = multiclass_nms(boxes_, num_classes=3, match_threshold=0.5)
    print(len(res))




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function:
# pip install scikit-learn
from collections import Counter

from sklearn.metrics import confusion_matrix as confusion_matrix_compute
from sklearn.metrics import fbeta_score, classification_report


class FBetaScore(object):

    def __init__(self, y_true: list, y_pre: list, f_beta: list, fbeta_score_average=None, report_label=None,
                 report_digits=4, zero_division=0):
        """

        :param y_true:
        :param y_pre:
        :param f_beta: [1, 1.5, 2]
        :param fbeta_score_average:  {'micro', 'macro', 'samples', 'weighted', 'binary'} or None, default='binary'
                This parameter is required for multiclass/multilabel targets.
                If ``None``, the scores for each class are returned. Otherwise, this
                determines the type of averaging performed on the data:

                ``'binary'``:
                    Only report results for the class specified by ``pos_label``.
                    This is applicable only if targets (``y_{true,pred}``) are binary.
                ``'micro'``:
                    Calculate metrics globally by counting the total true positives,
                    false negatives and false positives.
                ``'macro'``:
                    Calculate metrics for each label, and find their unweighted
                    mean.  This does not take label imbalance into account.
                ``'weighted'``:
                    Calculate metrics for each label, and find their average weighted
                    by support (the number of true instances for each label). This
                    alters 'macro' to account for label imbalance; it can result in an
                    F-score that is not between precision and recall.
                ``'samples'``:
                    Calculate metrics for each instance, and find their average (only
                    meaningful for multilabel classification where this differs from
                    :func:`accuracy_score`).
        :param report_label: 需要统计的类别，例如(1, 2, 3)
        :param report_digits: 统计指标保留的小数位数
        :param zero_division: "warn", 0 or 1, default="warn"
                              Sets the value to return when there is a zero division. If set to
                              "warn", this acts as 0, but warnings are also raised.
        """
        assert len(y_true) > 0, "y_true is empty."
        assert len(y_pre) > 0, "y_pre is empty."
        self.y_true = y_true
        self.y_pre = y_pre
        self.f_beta = f_beta if f_beta is not None else [1]
        self.fbeta_score_average = fbeta_score_average
        self.report_label = report_label
        self.report_digits = report_digits
        self.zero_division = zero_division
        self.tp_multi = {}
        self.tn_multi = {}
        self.fp_multi = {}
        self.fn_multi = {}
        self.label_count = Counter(self.y_true)
        self.tn, self.fp, self.fn, self.tp = self.confusion_matrix(y_true=self.y_true, y_pre=self.y_pre)
        self.recall = self.tp / (self.tp + self.fn) if (self.tp + self.fn) > 0 else 0
        self.recall = f'{self.recall:.{self.report_digits}f}'
        self.precision = self.tp / (self.tp + self.fp) if (self.tp + self.fp) > 0 else 0
        self.precision = f'{self.precision:.{self.report_digits}f}'
        self.f_beta_score = self.f_beta_score_compute(f_beta=self.f_beta, y_true=self.y_true, y_pre=self.y_pre)
        self.accuracy = (self.tp + self.tn) / (self.tp + self.tn + self.fp + self.fn)
        self.report = classification_report(self.y_true, self.y_pre, digits=self.report_digits, labels=self.report_label,
                                            zero_division=self.zero_division)

    def print(self, is_print=True):
        from prettytable import PrettyTable
        table = PrettyTable(field_names=["confusion matrix", "recall", "precision", "f_beta_score"],
                            title="F_beta Score Statistical Table (total={0})".format(len(self.y_true)))
        row = ["tp={0}  fp={1}\nfn={2}  tn={3}".format(self.tp, self.fp, self.fn, self.tn),
               "tp/(tp+fn)={0}/{1}\nrecall={2}".format(self.tp, self.tp + self.fn, self.recall),
               "tp/(tp+fp)={0}/{1}\nprecision={2}".format(self.tp, self.tp + self.fp, self.precision)]
        union = []
        for i, beta in enumerate(self.f_beta):
            union.append("f-{0}-score: {1}\n".format(beta, self.f_beta_score[i]))
        row.append("\n".join(union))
        table.add_row(row)
        table.align = "l"
        # table.align["confusion matrix"] = "c"
        if is_print:
            union = self.tn_multi | self.tn_multi | self.fp_multi | self.fn_multi
            if len(union) > 0:
                print("True Positives (TP):", self.tp_multi)
                print("True Negatives (TN):", self.tn_multi)
                print("False Positives (FP):", self.fp_multi)
                print("False Negatives (FN):", self.fn_multi)
            print(table)
            print(self.report)
        return table, self.report

    def confusion_matrix(self, y_true: list, y_pre: list):
        """
        1为正例，0为负例
        -------------
        TP 真正例，实际为1，预测为1，正样本预测正确
        FN 假负例，实际为1，预测为0，正样本预测错误
        -------------
        FP 假正例，实际为0，预测为1，负样本预测错误
        TN 真负例，实际为0，预测为0，负样本预测正确

        +------------------+------------------------------------ +
        |                  |              True Class             |
        | Predicted Class  |------------------+------------------+
        |                  |     Positive     |     Negative     |
        +------------------+------------------+------------------+
        |        Y         |        TP        |        FP        |
        +------------------+------------------+------------------+
        |        N         |        FN        |        TN        |
        +------------------+------------------+------------------+
        查全率： Recall = TP/(TP+FN)；     实际为正例的样本中，模型正确预测为正例的比例
        查准率： Precision = TP/(TP+FP)；  模型预测为正例的样本中，实际为正例的比例
        """
        if len(set(y_true+y_pre)) == 2:
            # 二分类
            self.tn, self.fp, self.fn, self.tp = confusion_matrix_compute(y_true=y_true, y_pred=y_pre).ravel()
        else:
            # 多分类
            cm = confusion_matrix_compute(y_true=y_true, y_pred=y_pre)
            # 遍历每个类别
            for i in range(len(cm)):
                self.tp_multi[i] = cm[i, i]  # 真正例
                self.fn_multi[i] = cm[i, :].sum() - cm[i, i]  # 假负例
                self.fp_multi[i] = cm[:, i].sum() - cm[i, i]  # 假正例
                self.tn_multi[i] = cm.sum() - (self.tp_multi[i] + self.fn_multi[i] + self.fp_multi[i])  # 真负例需要计算除了当前类别i之外的所有其他类别的TN之和
            self.tp = sum([v for k, v in self.tp_multi.items()])
            self.tn = sum([v for k, v in self.tn_multi.items()])
            self.fp = sum([v for k, v in self.fp_multi.items()])
            self.fn = sum([v for k, v in self.fn_multi.items()])
        return self.tn, self.fp, self.fn, self.tp

    def f_beta_score_compute(self, f_beta: list, y_true: list = None, y_pre: list = None):
        """
        Fβ分数 = (1 + β^2) * (精确率 * 召回率) / (β^2 * 精确率 + 召回率)
        β参数决定了精确率和召回率的相对权重
        当 β = 0 时，则Fβ分数等于precision
        当 β < 1 时，则模型偏向精确率
        当 β = 1 时，则Fβ=F1，精确率权重等于召回率权重
        当 β > 1 时，则模型偏向召回率
        :param f_beta:
        :param y_true:
        :param y_pre:
        """
        if y_true is None:
            y_true = self.y_true
        if y_pre is None:
            y_pre = self.y_pre
        f_score = []
        for beta in f_beta:
            f_print = "\n"
            scores = fbeta_score(y_true=y_true, y_pred=y_pre, beta=beta, average=self.fbeta_score_average,
                                 labels=self.report_label, zero_division=self.zero_division).tolist()
            if self.report_label is not None:
                report_label = list(self.report_label)
                for i, v in enumerate(report_label):
                    f_print += f'  {report_label[i]}: {scores[i]:.{self.report_digits}f}'
                    if i < len(report_label) - 1:
                        f_print += "\n"
                if self.report_label is not None:
                    micro_avg = fbeta_score(y_true=y_true, y_pred=y_pre, beta=beta, average="micro", labels=self.report_label, zero_division=self.zero_division)
                    macro_avg = fbeta_score(y_true=y_true, y_pred=y_pre, beta=beta, average="macro", labels=self.report_label, zero_division=self.zero_division)
                    weighted_avg = fbeta_score(y_true=y_true, y_pred=y_pre, beta=beta, average="weighted", labels=self.report_label, zero_division=self.zero_division)
                    f_print += f"\n  --------------------"
                    f_print += f"\n  micro avg: {micro_avg:.{self.report_digits}f}"
                    f_print += f"\n  macro avg: {macro_avg:.{self.report_digits}f}"
                    f_print += f"\n  weighted avg: {weighted_avg:.{self.report_digits}f}"
                f_score.append(f_print)
            else:
                f_score.append(fbeta_score(y_true=y_true, y_pred=y_pre, beta=beta, average=self.fbeta_score_average,
                                           labels=self.report_label, zero_division=self.zero_division))
                pass
        # f_score = (1 + beta ** 2) * (precision * recall) / (beta ** 2 * precision + recall)
        return f_score


if __name__ == "__main__":
    # actual_labels =    [0, 1, 0, 1, 1, 0, 0, 0, 0, 0]
    # predicted_labels = [0, 1, 0, 1, 0, 1, 1, 0, 0, 0]
    # actual_labels =    [0, 0, 0, 0, 0, 0, 1]
    # predicted_labels = [0, 0, 0, 0, 1, 1, 1]
    # actual_labels =    [-1, 1, -1, 2, 2, 3]
    # predicted_labels = [-1, 1, -1, 2, 1, 3]
    # actual_labels =    [1, 0, 2, 1, 0, 0]
    # predicted_labels = [1, 0, 2, 0, 1, 1]
    actual_labels =    ['1', '3', '2', '1_2', '2', '2']
    predicted_labels = ['1_2', '3', '2', '1', '1', '1']
    score = FBetaScore(y_true=actual_labels, y_pre=predicted_labels, f_beta=[0.5, 1], fbeta_score_average=None, report_label=("1", "2", "3", "4"))
    score.print()





# -*- coding: utf-8 -*-
# @Author  : zhousf-a
# @Function:
"""
pip install chromadb
pip3 install typing_extensions

eg:

import numpy as np
from chromadb.api.types import Documents, Embeddings, EmbeddingFunction
from chromadb.api.collection_configuration import (
            CreateCollectionConfiguration,
            CreateHNSWConfiguration
        )

class MyEmbeddingFunction(EmbeddingFunction):

    def __init__(self):
        pass

    def __call__(self, doc: Documents) -> Embeddings:
        embeddings = []
        demo = {
            "苹果": np.asarray([0.1, 0.2, 0.3]),
            "香蕉": np.asarray([0.2, 0.1, 0.4]),
            "橘子": np.asarray([0.15, 0.22, 0.35]),
            "橙子": np.asarray([0.15, 0.22, 0.5]),
        }
        for item in doc:
            embeddings.append(demo.get(item))
        return embeddings


master = ChromaDBMaster().create_collection(name="my_collection",
                                            configuration=CreateCollectionConfiguration(hnsw=CreateHNSWConfiguration(space="ip")),
                                            embedding_function=MyEmbeddingFunction())
master.add(
    ids=["id1", "id2", "id3"],
    documents=["苹果", "香蕉", "橘子"],
)
results = master.query(
    query_texts=["橙子"],
    n_results=3
)
print(results)
"""
from pathlib import Path
from typing import Union, Optional
from chromadb.api.models.Collection import Collection
import chromadb.utils.embedding_functions as ef
from chromadb.api.collection_configuration import (
    CreateCollectionConfiguration,
)
from chromadb.config import Settings

import chromadb
from chromadb.api.types import (
    Embeddable,
    EmbeddingFunction,
    DataLoader,
    Loadable,
    URI,
    CollectionMetadata,
    Embedding,
    PyEmbedding,
    Metadata,
    Document,
    Image,
    Where,
    IDs,
    GetResult,
    QueryResult,
    ID,
    OneOrMany,
    WhereDocument,
)


class ChromaDBMaster:
    def __init__(self, path: Union[str, Path] = None, settings: Optional[Settings] = None):
        """

        :param path: 保存数据库的目录，非空时则为持久化数据库，eg: .chromadb
        """
        if settings is None:
            settings = Settings(anonymized_telemetry=False)
        self.chroma_client = chromadb.PersistentClient(path=path, settings=settings) if path else chromadb.Client(settings=settings)
        self.collection = None

    def create_collection(self, name: str,
                          configuration: Optional[CreateCollectionConfiguration] = None,
                          metadata: Optional[CollectionMetadata] = None,
                          embedding_function: Optional[
                              EmbeddingFunction[Embeddable]
                          ] = ef.DefaultEmbeddingFunction(),
                          data_loader: Optional[DataLoader[Loadable]] = None,
                          get_or_create: bool = False) -> Collection:
        """Create a new collection with the given name and metadata.
               Args:
                   name: The name of the collection to create.
                   configuration:
                        {"hnsw": {"space": "ip"}}  内积
                        {"hnsw": {"space": "cosine"}} 余弦距离
                        {"hnsw": {"space": "l2"}} 欧式距离
                   metadata: Optional metadata to associate with the collection.
                   embedding_function: Optional function to use to embed documents.
                                       Uses the default embedding function if not provided.
                   get_or_create: If True, return the existing collection if it exists.
                   data_loader: Optional function to use to load records (documents, images, etc.)

               Returns:
                   Collection: The newly created collection.

               Raises:
                   ValueError: If the collection already exists and get_or_create is False.
                   ValueError: If the collection name is invalid.

               Examples:
                   ```python
                   client.create_collection("my_collection")
                   # collection(name="my_collection", metadata={})

                   client.create_collection("my_collection", metadata={"foo": "bar"})
                   # collection(name="my_collection", metadata={"foo": "bar"})
                   ```
               """
        params = locals().copy()
        params.pop("self")
        self.collection = self.chroma_client.create_collection(**params)
        return self.collection

    def count(self) -> int:
        """The total number of embeddings added to the database

        Returns:
            int: The total number of embeddings added to the database

        """
        return self.collection.count()

    def add(self,
            ids: OneOrMany[ID] = None,
            embeddings: Optional[Union[OneOrMany[Embedding], OneOrMany[PyEmbedding]]] = None,
            metadatas: Optional[OneOrMany[Metadata]] = None,
            documents: Optional[OneOrMany[Document]] = None,
            images: Optional[OneOrMany[Image]] = None,
            uris: Optional[OneOrMany[URI]] = None) -> None:
        """Add embeddings to the data store.
        Args:
            ids: The ids of the embeddings you wish to add
            embeddings: The embeddings to add. If None, embeddings will be computed based on the documents or images using the embedding_function set for the Collection. Optional.
            metadatas: The metadata to associate with the embeddings. When querying, you can filter on this metadata. Optional.
            documents: The documents to associate with the embeddings. Optional.
            images: The images to associate with the embeddings. Optional.
            uris: The uris of the images to associate with the embeddings. Optional.

        Returns:
            None

        Raises:
            ValueError: If you don't provide either embeddings or documents
            ValueError: If the length of ids, embeddings, metadatas, or documents don't match
            ValueError: If you don't provide an embedding function and don't provide embeddings
            ValueError: If you provide both embeddings and documents
            ValueError: If you provide an id that already exists

        """
        if ids is None:
            index = self.count()
            if documents:
                ids = [str(index + i) for i in range(1, len(documents) + 1)]
            if embeddings:
                ids = [str(index + i) for i in range(1, len(embeddings) + 1)]
        params = locals().copy()
        params.pop("self")
        if "index" in params:
            params.pop("index")
        self.collection.add(**params)

    def query(
            self,
            query_embeddings: Optional[
                Union[
                    OneOrMany[Embedding],
                    OneOrMany[PyEmbedding],
                ]
            ] = None,
            query_texts: Optional[OneOrMany[Document]] = None,
            query_images: Optional[OneOrMany[Image]] = None,
            query_uris: Optional[OneOrMany[URI]] = None,
            ids: Optional[OneOrMany[ID]] = None,
            n_results: int = 10,
            where: Optional[Where] = None,
            where_document: Optional[WhereDocument] = None,
            include=None,
    ) -> QueryResult:
        """Get the n_results nearest neighbor embeddings for provided query_embeddings or query_texts.

        Args:
            query_embeddings: The embeddings to get the closes neighbors of. Optional.
            query_texts: The document texts to get the closes neighbors of. Optional.
            query_images: The images to get the closes neighbors of. Optional.
            query_uris: The URIs to be used with data loader. Optional.
            ids: A subset of ids to search within. Optional.
            n_results: The number of neighbors to return for each query_embedding or query_texts. Optional.
            where: A Where type dict used to filter results by. E.g. `{"$and": [{"color" : "red"}, {"price": {"$gte": 4.20}}]}`. Optional.
            where_document: A WhereDocument type dict used to filter by the documents. E.g. `{"$contains": "hello"}`. Optional.
            include: A list of what to include in the results. Can contain `"embeddings"`, `"metadatas"`, `"documents"`, `"distances"`. Ids are always included. Defaults to `["metadatas", "documents", "distances"]`. Optional.

        Returns:
            QueryResult: A QueryResult object containing the results.

        Raises:
            ValueError: If you don't provide either query_embeddings, query_texts, or query_images
            ValueError: If you provide both query_embeddings and query_texts
            ValueError: If you provide both query_embeddings and query_images
            ValueError: If you provide both query_texts and query_images

        """
        if include is None:
            include = ["metadatas", "documents", "distances"]
        params = locals().copy()
        params.pop("self")
        return self.collection.query(**params)

    def get(
            self,
            ids: Optional[OneOrMany[ID]] = None,
            where: Optional[Where] = None,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            where_document: Optional[WhereDocument] = None,
            include=None,
    ) -> GetResult:
        """Get embeddings and their associate data from the data store. If no ids or where filter is provided returns
        all embeddings up to limit starting at offset.

        Args:
            ids: The ids of the embeddings to get. Optional.
            where: A Where type dict used to filter results by. E.g. `{"$and": [{"color" : "red"}, {"price": {"$gte": 4.20}}]}`. Optional.
            limit: The number of documents to return. Optional.
            offset: The offset to start returning results from. Useful for paging results with limit. Optional.
            where_document: A WhereDocument type dict used to filter by the documents. E.g. `{"$contains": "hello"}`. Optional.
            include: A list of what to include in the results. Can contain `"embeddings"`, `"metadatas"`, `"documents"`. Ids are always included. Defaults to `["metadatas", "documents"]`. Optional.

        Returns:
            GetResult: A GetResult object containing the results.

        """
        if include is None:
            include = ["metadatas", "documents"]
        params = locals().copy()
        params.pop("self")
        return self.collection.get(**params)

    def update(
            self,
            ids: OneOrMany[ID],
            embeddings: Optional[
                Union[
                    OneOrMany[Embedding],
                    OneOrMany[PyEmbedding],
                ]
            ] = None,
            metadatas: Optional[OneOrMany[Metadata]] = None,
            documents: Optional[OneOrMany[Document]] = None,
            images: Optional[OneOrMany[Image]] = None,
            uris: Optional[OneOrMany[URI]] = None,
    ) -> None:
        """Update the embeddings, metadatas or documents for provided ids.

        Args:
            ids: The ids of the embeddings to update
            embeddings: The embeddings to update. If None, embeddings will be computed based on the documents or images using the embedding_function set for the Collection. Optional.
            metadatas:  The metadata to associate with the embeddings. When querying, you can filter on this metadata. Optional.
            documents: The documents to associate with the embeddings. Optional.
            images: The images to associate with the embeddings. Optional.
            uris: The uris of the images to associate with the embeddings. Optional.
        Returns:
            None
        """
        params = locals().copy()
        params.pop("self")
        self.collection.update(**params)

    def upsert(
            self,
            ids: OneOrMany[ID],
            embeddings: Optional[
                Union[
                    OneOrMany[Embedding],
                    OneOrMany[PyEmbedding],
                ]
            ] = None,
            metadatas: Optional[OneOrMany[Metadata]] = None,
            documents: Optional[OneOrMany[Document]] = None,
            images: Optional[OneOrMany[Image]] = None,
            uris: Optional[OneOrMany[URI]] = None,
    ) -> None:
        """Update the embeddings, metadatas or documents for provided ids, or create them if they don't exist.

        Args:
            ids: The ids of the embeddings to update
            embeddings: The embeddings to add. If None, embeddings will be computed based on the documents using the embedding_function set for the Collection. Optional.
            metadatas:  The metadata to associate with the embeddings. When querying, you can filter on this metadata. Optional.
            documents: The documents to associate with the embeddings. Optional.
            images: The images to associate with the embeddings. Optional.
            uris: The uris of the images to associate with the embeddings. Optional.
        Returns:
            None
        """
        params = locals().copy()
        params.pop("self")
        self.collection.upsert(**params)

    def delete(
            self,
            ids: Optional[IDs] = None,
            where: Optional[Where] = None,
            where_document: Optional[WhereDocument] = None,
    ) -> None:
        """Delete the embeddings based on ids and/or a where filter

        Args:
            ids: The ids of the embeddings to delete
            where: A Where type dict used to filter the delection by. E.g. `{"$and": [{"color" : "red"}, {"price": {"$gte": 4.20}]}}`. Optional.
            where_document: A WhereDocument type dict used to filter the deletion by the document content. E.g. `{"$contains": "hello"}`. Optional.

        Returns:
            None

        Raises:
            ValueError: If you don't provide either ids, where, or where_document
        """
        params = locals().copy()
        params.pop("self")
        self.collection.delete(**params)




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: lmdb (Lightning Memory-Mapped Database) 快如闪电的内存映射数据库
# 支持多进程
"""
pip install lmdb
pip install msgpack

[question]
MDB_MAP_FULL: Environment mapsize limit reached
[solution]
提升map_size或定时删除历史数据

[question]
lmdb.PanicError: mdb_txn_begin: MDB_PANIC: Update of meta page failed or environment had fatal error

"""
import logging
import threading
from pathlib import Path
from typing import Optional, Union, Any, Dict

import lmdb
try:
    import msgpack
except ImportError:
    raise ImportError('Please install msgpack')


class LMDB(object):

    def __init__(self, db_dir: Union[str, Path],
                 map_size: int = int(1e9),
                 max_readers: int = 126,
                 max_dbs: int = 0,
                 readonly: bool = False):
        """
        :param db_dir: 数据库文件路径
        :param map_size: 数据库最大大小（字节）
                         1e9 ≈ 1GB;
                         1e10 ≈ 10GB;
                         1e11 ≈ 100GB(最大值)
        :param max_readers: 最大读取器数量
        :param max_dbs: 最大数据库数量，0表示使用默认值
        :param readonly: 是否只读模式
        """
        self.db_dir = Path(db_dir) if isinstance(db_dir, str) else db_dir
        if not self.db_dir.exists():
            self.db_dir.mkdir(parents=True, exist_ok=True)
        self.map_size = map_size
        self.max_readers = max_readers
        self.max_dbs = max_dbs
        self.readonly = readonly
        self._lock = threading.RLock()
        self.env = self.initialize()

    def initialize(self):
        try:
            return lmdb.open(
                path=str(self.db_dir),
                map_size=self.map_size,
                max_readers=self.max_readers,
                max_dbs=self.max_dbs,
                readonly=self.readonly,
                lock=True
            )
        except Exception as e:
            logging.error(f"initialize lmdb failed: {e}")
            raise

    @staticmethod
    def _encode_key(key: Union[str, bytes]) -> bytes:
        if isinstance(key, str):
            return key.encode('utf-8')
        return key

    @staticmethod
    def _decode_key(key: bytes) -> str:
        return key.decode('utf-8')

    @staticmethod
    def _encode_value(value: Any) -> bytes:
        return msgpack.packb(value, use_bin_type=True)

    @staticmethod
    def _decode_value(value_bytes: bytes) -> Any:
        return msgpack.unpackb(value_bytes, raw=False)

    def count(self) -> int:
        with self._lock:
            try:
                with self.env.begin() as txn:
                    return txn.stat()['entries']
            except Exception as e:
                logging.error(f"items count failed from lmdb: {e}")
                return 0

    def insert(self, key: Union[str, bytes], value: Any, txn: Optional[lmdb.Transaction] = None) -> bool:
        try:
            encoded_key = self._encode_key(key)
            encoded_value = self._encode_value(value)
            if txn:
                txn.put(encoded_key, encoded_value)
            else:
                with self.env.begin(write=True) as txn:
                    txn.put(encoded_key, encoded_value)
            return True
        except Exception as e:
            logging.error(f"item insert failed from lmdb: key: {key}, error: {e}")
            return False

    def delete(self, key: Union[str, bytes], txn: Optional[lmdb.Transaction] = None) -> bool:
        with self._lock:
            try:
                encoded_key = self._encode_key(key)
                if txn:
                    txn.delete(encoded_key)
                else:
                    with self.env.begin(write=True) as txn:
                        txn.delete(encoded_key)
                return True
            except Exception as e:
                logging.error(f"clear_all failed from lmdb: {e}")
                return False

    def update(self,  key: Union[str, bytes], value: Any, txn: Optional[lmdb.Transaction] = None) -> bool:
        return self.insert(key, value, txn)

    def exists(self, key: Union[str, bytes]) -> bool:
        with self._lock:
            try:
                encoded_key = self._encode_key(key)
                with self.env.begin() as txn:
                    return txn.get(encoded_key) is not None
            except Exception as e:
                logging.error(f"query failed from lmdb: key: {key}, error: {e}")
                return False

    def query(self, key: Union[str, bytes], txn: Optional[lmdb.Transaction] = None, default: Any = None) -> Any:
        with self._lock:
            try:
                encoded_key = self._encode_key(key)
                if txn:
                    value_bytes = txn.get(encoded_key)
                    if value_bytes is None:
                        return default
                    return self._decode_value(value_bytes)
                else:
                    with self.env.begin() as txn:
                        value_bytes = txn.get(encoded_key)
                        if value_bytes is None:
                            return default
                        return self._decode_value(value_bytes)
            except Exception as e:
                logging.error(f"item query failed from lmdb: {e}")
                return default

    def query_all(self) -> Dict[str, Any]:
        with self._lock:
            result = {}
            try:
                with self.env.begin() as txn:
                    cursor = txn.cursor()
                    for encoded_key, value_bytes in cursor:
                        key = self._decode_key(encoded_key)
                        value = self._decode_value(value_bytes)
                        result[key] = value
                return result
            except Exception as e:
                logging.error(f"query_all failed from lmdb: {e}")
            return result

    def display(self):
        with self._lock:
            try:
                with self.env.begin() as txn:
                    cursor = txn.cursor()
                    for encoded_key, value_bytes in cursor:
                        key = self._decode_key(encoded_key)
                        value = self._decode_value(value_bytes)
                        print(key, value)
            except Exception as e:
                logging.error(f"display failed from lmdb: {e}")

    def clear_all(self) -> bool:
        with self._lock:
            try:
                with self.env.begin(write=True) as txn:
                    cursor = txn.cursor()
                    for key, _ in cursor:
                        txn.delete(key)
                return True
            except Exception as e:
                logging.error(f"clear_all failed from lmdb: {e}")
                return False

    def backup(self, backup_path: Union[str, Path]) -> bool:
        with self._lock:
            try:
                if isinstance(backup_path, str):
                    backup_path = Path(backup_path)
                backup_path.parent.mkdir(parents=True, exist_ok=True)
                with self.env.begin() as txn:
                    with lmdb.open(str(backup_path), map_size=self.map_size) as backup_env:
                        with backup_env.begin(write=True) as backup_txn:
                            cursor = txn.cursor()
                            for encoded_key, value_bytes in cursor:
                                backup_txn.put(encoded_key, value_bytes)

                logging.info(f"database backup from lmdb success: {backup_path}")
                return True
            except Exception as e:
                logging.error(f"database backup from lmdb failed: {e}")
                return False

    def get_stats(self) -> Dict[str, Any]:
        with self._lock:
            try:
                with self.env.begin() as txn:
                    stats = txn.stat()
                    info = self.env.info()
                return {
                    'entries': stats['entries'],
                    'depth': stats['depth'],
                    'branch_pages': stats['branch_pages'],
                    'leaf_pages': stats['leaf_pages'],
                    'overflow_pages': stats['overflow_pages'],
                    'map_size': info['map_size'],
                    'last_txnid': info['last_txnid'],
                    'max_readers': info['max_readers'],
                    'num_readers': info['num_readers']
                }
            except Exception as e:
                logging.error(f"fetch database stats failed: {e}")
                return {}

    def close(self):
        with self._lock:
            try:
                if hasattr(self, 'env'):
                    self.env.close()
                logging.info("close database connection successfully.")
            except Exception as e:
                logging.error(f"close database connection failed: {e}")

    def __enter__(self):
        """上下文管理器入口"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """上下文管理器出口"""
        self.close()

    def __del__(self):
        """析构函数"""
        self.close()


if __name__ == '__main__':
    db = LMDB(Path(__file__).parent)
    try:
        # 测试不同类型的数据
        test_cases = [
            ("int_value", 12345),
            ("float_value", 123.456),
            ("string_value", "Hello, World!"),
            ("dict_value", {"name": "张三", "age": 25}),
            ("list_value", [1, 2, 3, "a", "b", "c"]),
            ("bytes_value", b"binary_data_here")
        ]
        for k, v in test_cases:
            # 写入数据
            if db.insert(k, v):
                print(f"  ✅ 写入数据成功: {k}")
            else:
                print(f"  ❌ 写入数据失败: {k}")

        print("\n测试更新功能:")
        print(f"更新数据: {db.update(key='int_value', value=1000)}")

        print("\n测试查询功能:")
        print(f"查询数据: {db.query_all()}")

        print("\n测试删除功能:")
        print(f"删除数据: {db.delete(key='int_value')}")

        # 测试统计信息
        stats = db.get_stats()
        print(f"\n数据库统计: {db.count()} 条记录")
        print(f"数据库统计详细信息: {stats}")

        # 测试搜索功能
        print("\n测试搜索功能:")
        print(f"搜索结果: {db.query(key='dict_value')}")

        # 测试删除所有数据功能
        print("\n测试删除所有数据功能:")
        print(f"删除所有数据: {db.clear_all()}")
    finally:
        db.close()





# -*- coding: utf-8 -*-
# @Author  : zhousf-a
# @Function:
"""
python -m pip install paddle2onnx
python -m pip install paddlepaddle-gpu==2.6.2 -i https://www.paddlepaddle.org.cn/packages/stable/cu118/
pip install --upgrade pyOpenSSL -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
pip install httpx -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
pip install sqlite3py -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
"""
import os
from pathlib import Path


def convert_onnx(inference_model_dir: Path, opset_version=13):
    save_onnx_file = inference_model_dir.joinpath("model.onnx")
    command = ("paddle2onnx --model_dir {0} "
               "--model_filename model.pdmodel "
               "--params_filename model.pdiparams "
               "--save_file {1} "
               "--opset_version {2} ").format(str(inference_model_dir), str(save_onnx_file), opset_version)
    os.system(command)




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Function: 相似度计算
import re
import numpy as np
import jieba
from typing import List
from zhousflib.util import re_util
from zhousflib.metrics.cosine import Cosine
from sklearn.metrics.pairwise import cosine_similarity
from zhousflib.ml.feature_vector import FeatureVector, TypeFeatureVector

"""
字符串匹配算法：这是最基本的文本相似度计算方法，主要通过将两个文本字符串进行逐个字符的比较，计算相同字符的数量占总字符数的比例来判断文本的相似度。但是，这种方法对于大量文本的比对速度较慢，且只能检测出完全相同的文本
哈希算法：哈希算法可以快速计算出文本的哈希值，然后通过比对哈希值来判断文本的相似度。但是，哈希算法存在哈希冲突的问题，即不同的文本可能会产生相同的哈希值，从而导致误判
N-gram算法：N-gram算法是一种基于文本分词的方法，将文本分成N个连续的词组，然后比对词组的相似度来判断文本的相似度。N-gram算法可以识别出部分相似的文本，相对于字符串匹配算法和哈希算法，其检测精度更高。
向量空间模型算法：向量空间模型算法是一种基于文本向量化的方法，将文本转换成向量，然后计算向量之间的相似度来判断文本的相似度。这种方法可以识别出语义相似的文本，相对于其他算法，其检测精度更高。

MinHash算法：MinHash算法是一种基于哈希的文本查重方法，它通过随机排列文档中的词项并使用哈希函数来比较文档的相似性。
SimHash算法：是一种局部敏感哈希（Locality Sensitive Hashing, LSH）技术，最初由Google提出，用于高效地计算文本相似度
。其核心思想是将相似的文本映射到相近的哈希空间中，从而实现快速的相似性检测。

运行速度：KSentence > Simhash > Minhash
准确率：KSentence > Minhash > Simhash
召回率：Simhash > Minhash > KSentence
工程应用上，海量文本用Simhash，短文本用Minhash，追求速度用KSentence。

余弦相似度：from sklearn.metrics.pairwise import cosine_similarity   
欧氏距离：  from sklearn.metrics.pairwise import euclidean_distances
曼哈顿距离：from sklearn.metrics.pairwise import manhattan_distances
"""


def text_to_vector(text: List[str], vector_type=TypeFeatureVector.TYPE_COUNT_VECTOR):
    return FeatureVector(vector_type=vector_type).fit_transform(text)


def compute_similarity_pairwise(text1: List[str], text2: List[str], vector_type=TypeFeatureVector.TYPE_COUNT_VECTOR,
                                filter_punctuation=True, filter_s=True,
                                cut_all=True, filter_result=False,
                                filter_threshold: float = 0):
    """
    计算两个文本列表之间的相似度矩阵
    :param text1: 第一个文本列表
    :param text2: 第二个文本列表
    :param vector_type:  特征向量类型，默认为TypeFeatureVector.TYPE_COUNT_VECTOR
    :param filter_punctuation: 是否过滤标点符号
    :param filter_s: 是否过滤空格、制表符、换行符
    :param cut_all: 是否全模式分词
    :param filter_result: 是否过滤结果
    :param filter_threshold: 过滤阈值
    :return: 相似度矩阵
    """
    def deal_text(text):
        _text = []
        for txt in text:
            if filter_punctuation:
                txt = re_util.remove_special_punctuation(str(txt))
                if filter_s:
                    # 提取中文、数字、字母
                    txt = re.sub(r"[^\u0041-\u005a\u0061-\u007a\u0030-\u0039\u4e00-\u9fa5]", "", txt)
                else:
                    # 提取中文、数字、字母、包括 \s （空格、制表符、换行符等）
                    txt = re.sub(r"[^\u0041-\u005a\u0061-\u007a\u0030-\u0039\u4e00-\u9fa5\s]", "", txt)
            _text.append(str(jieba.lcut(txt, cut_all=cut_all)))
        return text_to_vector(_text, vector_type)

    vector = deal_text(text1+text2)
    clip_index = len(text1)
    similarity_matrix = cosine_similarity(vector)
    if filter_threshold > 0:
        similarity_matrix = np.where(similarity_matrix < filter_threshold, 0, similarity_matrix)
    if filter_result:
        res = []
        clip_upper_triangular = similarity_matrix[:clip_index, clip_index:]
        nonzero = np.nonzero(clip_upper_triangular)
        for i in range(len(nonzero[0])):
            res.append((nonzero[0][i], nonzero[1][i] + clip_index, clip_upper_triangular[nonzero[0][i]][nonzero[1][i]]))
        return res
    return similarity_matrix


def compute_similarity(text: List[str], vector_type=TypeFeatureVector.TYPE_COUNT_VECTOR, filter_punctuation=True,
                       filter_s=True,
                       cut_all=True, filter_result=False, filter_threshold: float = 0):
    """

    :param text:
    :param vector_type:
    :param filter_punctuation: 过滤标点符号
    :param filter_s: 中文计算时，过滤空格、制表符、换行符，若计算英文时filter_s设置为False
    :param cut_all:
    :param filter_result:
    :param filter_threshold:
    :return:
    """
    _text = []
    for txt in text:
        if filter_punctuation:
            txt = re_util.remove_special_punctuation(str(txt))
            if filter_s:
                # 提取中文、数字、字母
                txt = re.sub(r"[^\u0041-\u005a\u0061-\u007a\u0030-\u0039\u4e00-\u9fa5]", "", txt)
            else:
                # 提取中文、数字、字母、包括 \s （空格、制表符、换行符等）
                txt = re.sub(r"[^\u0041-\u005a\u0061-\u007a\u0030-\u0039\u4e00-\u9fa5\s]", "", txt)
        _text.append(str(jieba.lcut(txt, cut_all=cut_all)))
    cosine = Cosine()
    vector = text_to_vector(_text, vector_type)
    similarity_matrix = cosine.cosine_vector_with_matrix(vector, filter_result=filter_result,
                                                         filter_threshold=filter_threshold)
    return similarity_matrix


if __name__ == "__main__":
    import time

    start = time.time()
    # documents = ["This is the first document",
    #              "This document is the second document",
    #              "This is the first book",
    #              "This document is the second book",
    #              "This is the third book"]
    # results = compute_similarity(text=documents, vector_type=TypeFeatureVector.TYPE_COUNT_VECTOR,
    #                              filter_punctuation=True, filter_s=False, cut_all=False, filter_result=True,
    #                              filter_threshold=0.01)
    # for it in results:
    #     print(it)
    # print("耗时", time.time() - start)

    results = compute_similarity_pairwise(["This is the first document", "This document is the second document"],
                                          ["This is the first book", "This document is the second book",
                                           "This is the third book"],
                                          vector_type=TypeFeatureVector.TYPE_COUNT_VECTOR,
                                          filter_punctuation=True, filter_s=False, cut_all=False, filter_result=True,
                                          filter_threshold=0.01)
    for it in results:
        print(it)
    print("耗时", time.time() - start)



"""
pip install python-docx
"""
import os
import re
import uuid
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from urllib.parse import urljoin
from typing import Union

from docx import Document
from docx.oxml.ns import qn
from docx.table import Table
from docx.table import _Cell
from docx.text.paragraph import Paragraph


class SectionType(str, Enum):
    HEADING = 'heading'
    PARAGRAPH = 'paragraph'
    TABLE = 'table'


@dataclass
class Section:
    type: SectionType
    content: Union[str, None] = None
    level: Union[str, None] = None
    contains_img: bool = False


class DocxParser:

    def __init__(self, docx_path: str, image_handler=None, preserve_style=True):
        self.docx_path = docx_path
        self.doc = Document(self.docx_path)
        if image_handler is None:
            image_handler = DocxParser.save_img_to_local
        self.image_handler = image_handler
        self.preserve_style = preserve_style

        self.para_number_set = defaultdict(list)
        self.numbering_part = {}

    def parse(self) -> list[Section]:
        sections = []
        self.para_number_set.clear()
        self.numbering_part = self.get_numbering_part(self.doc)
        for child in self.doc.element.body.iterchildren():
            if child.tag.endswith('p'):
                paragraph = Paragraph(child, self.doc)
                section = self.handle_paragraph(paragraph)
                if section:
                    sections.append(section)
            elif child.tag.endswith('tbl'):
                table = Table(child, self.doc)
                section = self.handle_table(table)
                sections.append(section)

        return sections

    def to_markdown(self) -> str:
        """
        Read the docx and parse it to markdown.
        :param docx_path:
        :param image_handler: The handler who responsible for storing images.
         The signature is image_handler(file_bytes, file_ext)
        :return: markdown
        """
        sections = self.parse()

        texts = []
        for section in sections:
            text = section.content
            if section.type != SectionType.HEADING:
                text += '\n'
            texts.append(text)
        return "\n".join(texts)  # md中，两个空格+换行才能表示换行

    def handle_paragraph(self, paragraph: Paragraph) -> Union[Section, None]:
        texts = []
        level = self.get_paragraph_heading_level(paragraph)
        type = SectionType.PARAGRAPH
        if level > 0:
            texts.append('#' * level + ' ')
            type = SectionType.HEADING
        numbering = self.get_paragraph_numbering(paragraph)
        if numbering:
            texts.append(numbering)

        # 空标题跳过，markdown解析器无法正确识别
        if not paragraph.runs and level > 0:
            return None

        buffer = []
        current_style = None

        def flush_text_buffer():
            if not buffer:
                return
            buffer_text = ''.join(buffer)
            if self.preserve_style and buffer_text.strip():
                if current_style == "bold":
                    prefix, core, suffix = self.split_text_by_whitespace(buffer_text)
                    buffer_text = f"{prefix}**{core}**{suffix}"
                elif current_style == "italic":
                    prefix, core, suffix = self.split_text_by_whitespace(buffer_text)
                    buffer_text = f"{prefix}*{core}*{suffix}"
            texts.append(buffer_text)
            buffer.clear()

        for run in paragraph.runs:
            if run.text:
                text = run.text
                style = "bold" if run.bold else "italic" if run.italic else "normal"
                if style != current_style:
                    flush_text_buffer()
                    current_style = style
                buffer.append(text)

            if run.element.drawing_lst:
                for drawing in run.element.drawing_lst:
                    image_data, ext = self.extract_image_bytes_and_ext(drawing, run.part)
                    if image_data:
                        flush_text_buffer()
                        img_path = self.image_handler(image_data, ext)
                        image_name = os.path.splitext(os.path.basename(img_path))[0]
                        texts.append(f'\n![{image_name}]({img_path})\n')

        flush_text_buffer()
        text = ''.join(texts)
        if not text.strip():  # 去掉空白行
            return None
        if re.fullmatch(r'\s*-{3,}\s*', text):  # 分隔线前面加个换行才会有效果
            text = '\n' + text
        return Section(type=type, content=text)

    @staticmethod
    def split_text_by_whitespace(text: str) -> tuple[str, str, str]:
        match = re.match(r'^(\s*)(.*?)(\s*)$', text, re.DOTALL)
        if match:
            return match.group(1), match.group(2), match.group(3)
        return '', text, ''

    @staticmethod
    def get_paragraph_heading_level(paragraph: Paragraph):
        if paragraph.style is not None:
            if paragraph.style.name.startswith('Heading'):
                # 使用样式名中的 "Heading X"
                return int(paragraph.style.name.split(' ')[-1])
            else:
                # 查找样式中的大纲级别
                style_element = paragraph.style._element
                pPr = style_element.find(qn('w:pPr'))
                if pPr is not None:
                    outlineLvl = pPr.find(qn('w:outlineLvl'))
                    if outlineLvl is not None:
                        outline_level = int(outlineLvl.get(qn('w:val')))
                        return outline_level + 1

        # 查找段落中的大纲级别
        pPr = paragraph._element.find(qn('w:pPr'))
        if pPr is not None:
            outlineLvl = pPr.find(qn('w:outlineLvl'))
            if outlineLvl is not None:
                outline_level = int(outlineLvl.get(qn('w:val')))
                return outline_level + 1

        return 0

    def get_paragraph_numbering(self, paragraph: Paragraph):
        pPr = paragraph._p.find(qn('w:pPr'))
        if pPr is None:
            return ''
        numPr = pPr.find(qn('w:numPr'))
        if numPr is not None:
            # 如果存在编号，取具体编号值（这里只是示例，真实编号格式需要结合 numbering.xml 判断）
            ilvl = numPr.find(qn('w:ilvl'))
            ilvl_val = ilvl.val if ilvl is not None else 0
            numId = numPr.find(qn('w:numId'))
            numId_val = numId.val if numId is not None else 0
            if numId_val > 0:
                para_number_arr = self.para_number_set[numId_val]
                while ilvl_val >= len(para_number_arr):
                    para_number_arr.append(0)
                para_num = para_number_arr[ilvl_val] + 1
                para_number_arr[ilvl_val] = para_num
                for i in range(ilvl_val + 1, len(para_number_arr)):
                    para_number_arr[i] = 0

                number_str = DocxNumberFormatter.generate_numbering(
                    para_number_arr[:ilvl_val + 1], self.numbering_part.get(str(numId_val), {})
                )
                if number_str:
                    return '    ' * ilvl_val + number_str + ' '
        return ''

    @staticmethod
    def extract_image_bytes_and_ext(drawing, part):
        """从 drawing XML 中提取图片数据及扩展名"""
        # 虑组合图片的场景
        blip = drawing.findall('.//' + qn('a:blip'))
        if blip:
            rid = blip[0].get(qn('r:embed'))
            image_part  = part.related_parts[rid]
            content_type = image_part .content_type  # 例如：image/png、image/jpeg
            ext = content_type.split('/')[-1]  # 提取扩展名部分，如 png、jpeg
            # 一些扩展名需要调整为标准格式（如 jpeg -> jpg）
            ext = {'jpeg': 'jpg'}.get(ext, ext)
            return image_part.blob, ext
        return None, None

    def handle_table(self, table: Table) -> Section:
        html = ['<table>']

        def get_cell_vMerge(tr_list, row_idx, col_idx):
            """获取指定单元格的纵向合并标志：continue代表是上一行单元格的合并部分"""
            if row_idx < len(tr_list):
                row = tr_list[row_idx]
                current_col_idx = 0
                for cell in row.tc_lst:
                    current_col_idx += cell.grid_span
                    if current_col_idx == col_idx:
                        return cell.vMerge
            return None

        for row_idx, row in enumerate(table._tbl.tr_lst):
            html.append('<tr>')
            col_idx = 0
            for tc in row.tc_lst:
                rowspan = 1
                colspan = tc.grid_span
                col_idx += colspan
                # 处理 vMerge（纵向合并 rowspan）
                if tc.vMerge == 'continue':
                    continue
                elif tc.vMerge == 'restart':
                    idx = row_idx + 1
                    while idx < len(table._tbl.tr_lst):
                        vMerge = get_cell_vMerge(table._tbl.tr_lst, idx, col_idx)
                        idx += 1
                        if vMerge == 'continue':
                            rowspan += 1
                        else:
                            break

                # 构造属性
                attr = []
                if rowspan > 1:
                    attr.append(f'rowspan="{rowspan}"')
                if colspan > 1:
                    attr.append(f'colspan="{colspan}"')

                # 构造临时 cell 获取内容
                cell = _Cell(tc, table)
                texts = []
                for paragraph in cell.paragraphs:
                    section = self.handle_paragraph(paragraph)
                    if section:
                        texts.append(section.content)
                content = '\n'.join(texts).strip()

                html.append(f'<td{(" " + " ".join(attr)) if attr else ""}>{content}</td>')
                col_idx += colspan
            html.append('</tr>')
        html.append('</table>')
        return Section(type=SectionType.TABLE, content=''.join(html))

    @staticmethod
    def get_numbering_part(doc):
        try:
            numbering_part_element = doc.part.numbering_part.element
        except NotImplementedError:
            return {}
        w_val = qn('w:val')
        # 1. Locate all w:num elements. Each w:num element contains a numId attribute that associates it with a paragraph.
        w_num = qn('w:num')
        num_elements = numbering_part_element.findall(w_num)

        # 2. Locate all abstractNumId elements. Typically, there is one abstractNumId element.
        absNumId_to_numId = defaultdict(list)
        w_abstractNumId = qn('w:abstractNumId')
        w_numId = qn('w:numId')
        for num_element in num_elements:
            abstractNumId = num_element.findall(w_abstractNumId)
            if len(abstractNumId) == 0:
                continue
            abstractNumId = abstractNumId[0]
            abstractNumId = abstractNumId.get(w_val)
            numId = num_element.get(w_numId)
            if abstractNumId is not None and numId is not None:
                absNumId_to_numId[abstractNumId].append(numId)

        # 3. Locate all abstractNum elements.
        w_abstractNum = qn('w:abstractNum')
        abstractNum_elements = numbering_part_element.findall(w_abstractNum)

        # 4. Within each abstractNum element, examine the abstractNumId, lvl, lvlText, and numFmt elements.
        # Under normal circumstances, there would be only one lvlText and one numFmt element for each level.
        w_lvl = qn('w:lvl')
        w_ilvl = qn('w:ilvl')
        w_lvlText = qn('w:lvlText')
        w_numFmt = qn('w:numFmt')

        numbering_part = {}
        for abstractNum_element in abstractNum_elements:
            abstractNumId = abstractNum_element.get(w_abstractNumId)
            if abstractNumId is None:
                continue
            bucket = {}
            lvl_elements = abstractNum_element.findall(w_lvl)
            numFmt = 'decimal'
            for lvl_element in lvl_elements:
                ilvl = lvl_element.get(w_ilvl)
                if ilvl is None:
                    continue

                lvlText_elements = lvl_element.findall(w_lvlText)
                numFmt_elements = lvl_element.findall(w_numFmt)
                if len(lvlText_elements) == 0 or len(numFmt_elements) == 0:
                    continue
                numFmt = numFmt_elements[0].get(w_val)
                text = lvlText_elements[0].get(w_val)
                bucket.update({ilvl: [text, numFmt]})

            if abstractNumId in absNumId_to_numId.keys():
                for numId in absNumId_to_numId[abstractNumId]:
                    numbering_part.update({
                        numId: bucket
                    })
        return numbering_part

    @staticmethod
    def save_img_to_local(image_bytes, ext, prefix='images'):
        os.makedirs(prefix, exist_ok=True)
        file_name = f"{uuid.uuid4().hex}.{ext}"
        img_path = urljoin(prefix + '/', file_name)
        with open(img_path, 'wb') as f:
            f.write(image_bytes)
        return img_path

    @classmethod
    def _run_contains_chart(cls, run) -> bool:
        """
        判断一个 Word run 是否包含图表（chart）
        """
        if run.element.drawing_lst:
            for drawing in run.element.drawing_lst:
                graphic_data_list = drawing.findall('.//' + qn('a:graphicData'))
                for graphic_data in graphic_data_list:
                    # 通过 tag 判断是否为 chart
                    for child in graphic_data:
                        if child.tag.endswith('chart') or child.tag.endswith('diagram') \
                                or child.tag.endswith('oleobject') or child.tag.endswith('grpSp'):
                            return True
        return False

    def contains_chart(self):
        for para in self.doc.paragraphs:
            for run in para.runs:
                if self._run_contains_chart(run):
                    return True
        return False


class DocxNumberFormatter:

    @staticmethod
    def int_to_lower_letter(n):
        """1 -> a, 2 -> b, ... 27 -> aa"""
        result = ""
        while n > 0:
            n -= 1
            result = chr(97 + n % 26) + result
            n //= 26
        return result

    @staticmethod
    def int_to_lower_roman(n):
        """1 -> i, 2 -> ii, etc."""
        val_map = [
            (1000, 'm'), (900, 'cm'), (500, 'd'), (400, 'cd'),
            (100, 'c'), (90, 'xc'), (50, 'l'), (40, 'xl'),
            (10, 'x'), (9, 'ix'), (5, 'v'), (4, 'iv'), (1, 'i')
        ]
        result = ""
        for val, sym in val_map:
            while n >= val:
                result += sym
                n -= val
        return result

    @staticmethod
    def get_styled_value(val: int, style: str) -> str:
        if style == 'decimal':
            return str(val)
        elif style == 'lowerLetter':
            return DocxNumberFormatter.int_to_lower_letter(val)
        elif style == 'upperLetter':
            return DocxNumberFormatter.int_to_lower_letter(val).upper()
        elif style == 'lowerRoman':
            return DocxNumberFormatter.int_to_lower_roman(val)
        elif style == 'upperRoman':
            return DocxNumberFormatter.int_to_lower_roman(val).upper()
        return str(val)

    @staticmethod
    def generate_numbering(level_values, numbering_map):
        """
        level_values: [1, 2, 3] 表示各层级序号值（从1开始）
        numbering_map: dict，层级样式定义
        """
        level = str(len(level_values) - 1)
        if level not in numbering_map:
            return ''
        fmt, style = numbering_map[level]

        # 遍历格式中的 %1 ~ %9，并替换为带格式的值
        for i, val in enumerate(level_values):
            placeholder = f'%{i + 1}'
            if placeholder in fmt:
                fmt = fmt.replace(placeholder, DocxNumberFormatter.get_styled_value(val, style))

        return fmt


if __name__ == "__main__":
    docx_path = r'C:\Users\zhousf-a\Desktop\图片敏感信息识别技术方案.docx'
    parser = DocxParser(docx_path)
    sections = parser.parse()
    for section in sections:
        print(f"Type: {section.type}, Content: {section.content}, Level: {section.level}, Contains Image: {section.contains_img}")

    markdown = parser.to_markdown()
    print(markdown)

    if parser.contains_chart():
        print("The document contains charts.")
    else:
        print("No charts found in the document.")
    # 如果需要将markdown保存到文件
    with open('output.md', 'w', encoding='utf-8') as f:
        f.write(markdown)
    print("Markdown content saved to output.md")



# -*- coding:utf-8 -*-
# Author:  zhousf
# Description:
import cv2
import numpy as np
from typing import Union
from pathlib import Path

"""
图像的高频信息、低频信息
低频信息：代表着图像中亮度/灰度值/颜色变化很缓慢的区域，描述了图像的主要部分，是对整幅图像强度的综合度量
高频信息：对应着图像变化剧烈的部分，也就是图像的边缘/轮廓、噪声以及细节部分，主要是对图像边缘/轮廓的度量，而人眼对高频分量比较敏感
"""


def write(image: np.ndarray, img_write_path: Path):
    """
    写图片-兼容图片路径包含中文
    :param image:
    :param img_write_path:
    :return:
    """
    cv2.imencode(img_write_path.suffix, image[:, :, ::-1])[1].tofile(str(img_write_path))


def read(img_path: Union[str, Path, np.ndarray], bg_to_white=False, overwrite=False) -> np.ndarray:
    """
    读图片-兼容图片路径包含中文
    :param img_path:
    :param bg_to_white: 是否将图片的透明背景转成白色背景
    :param overwrite: 是否覆盖原图，如果为True，则会将处理后的图片覆盖原图
    :return: np.ndarray
    """
    if isinstance(img_path, str):
        img_path = Path(img_path)
    if isinstance(img_path, Path):
        if bg_to_white:
            img_arr = cv2.imdecode(np.fromfile(img_path, dtype=np.uint8), cv2.IMREAD_UNCHANGED)
            if img_arr is not None and img_arr.ndim > 2 and img_arr.shape[2] == 4:
                alpha_channel = img_arr[:, :, :3].copy()
                # 找到非255且非0的像素点（半透明像素）
                transparent_mask = (alpha_channel != 255) & (alpha_channel != 0)
                if np.any(transparent_mask):
                    # 将透明像素设为白色
                    alpha_channel[transparent_mask] = [255, 255, 255]
                    img_arr = alpha_channel
                if overwrite:
                    write(img_arr, img_path)
        else:
            img_arr = cv2.imdecode(np.fromfile(str(img_path), dtype=np.uint8), cv2.IMREAD_COLOR)
        return img_arr
    return img_path


def is_transparent(img_path: Path):
    """
    判断图片是否透明背景
    :param img_path:
    :return:
    """
    img = cv2.imdecode(np.fromfile(str(img_path), dtype=np.uint8), cv2.IMREAD_UNCHANGED)
    if img is not None and img.shape[2] == 4:
        alpha_channel = img[:, :, 3]
        if cv2.countNonZero(alpha_channel) < alpha_channel.size:
            return True
    return False


def transparent_bg_to_white(img_path: Path, save_path: Path = None):
    """
    图片透明背景转成白色背景
    :param img_path:
    :param save_path:
    :return:
    """
    img = cv2.imdecode(np.fromfile(img_path, dtype=np.uint8), cv2.IMREAD_UNCHANGED)
    if img is not None and len(img.shape) > 2 and img.shape[2] == 4:
        alpha_channel = img[:, :, 3]
        if cv2.countNonZero(alpha_channel) < alpha_channel.size:
            b_channel, g_channel, r_channel, a_channel = cv2.split(img)
            white_background = np.ones_like(a_channel) * 255
            a_channel = a_channel / 255.0
            r_channel = r_channel * a_channel + white_background * (1 - a_channel)
            g_channel = g_channel * a_channel + white_background * (1 - a_channel)
            b_channel = b_channel * a_channel + white_background * (1 - a_channel)
            result = cv2.merge((b_channel, g_channel, r_channel))
            if save_path is not None:
                write(result, save_path)
            return result
    return img


def is_pure_color(image: Union[str, Path, np.ndarray], bg_to_white=False, overwrite=False):
    """
    判断图片是否为纯色
    :param image: 可以是图片路径或numpy数组
    :param bg_to_white: 是否将背景色转换为白色
    :param overwrite: 是否覆盖原图，如果为True，则会将处理后的图片覆盖原图
    """
    if not isinstance(image, np.ndarray):
        image = read(image, bg_to_white=bg_to_white, overwrite=overwrite)
    """
    获取图片左上角第一个像素点的颜色值: image[0, 0]，判断整个图片的颜色值是否都与左上角第一个像素点的颜色值相同
    例如：如果图片是纯白色，则image[0, 0]的值为[255, 255, 255]，如果整个图片都是纯白色，则np.all(image[0, 0] == image)将返回True
    这种方法的时间复杂度为O(n)，其中n为图片的像素数量
    这种方法的空间复杂度为O(1)，因为只需要存储一个像素点的颜色值
    优势：简单高效，时间复杂度低，空间复杂度低
    局限性：依赖于左上角像素作为参考，如果左上角像素恰好是噪声点，可能误判，对于有轻微噪声的图片可能不够鲁棒
    """
    return np.all(image[0, 0] == image)


def get_image_colors(image: Union[str, Path, np.ndarray], bg_to_white=False) -> int:
    """
    获取图片的颜色数量
    :param image: 可以是图片路径或numpy数组
    :param bg_to_white: 是否将背景色转换为白色
    :return: int
    """
    image_arr = read(image, bg_to_white)
    if image_arr.ndim == 2:  # 灰度图
        return len(np.unique(image_arr))
    elif image_arr.ndim == 3:  # 彩色图
        colors = np.unique(image_arr.reshape(-1, image_arr.shape[-1]), axis=0)
        return len(colors)
    else:
        return 0

